import logging
from langchain_core.prompts import ChatPromptTemplate
import time
from langchain.agents import AgentExecutor, create_openai_functions_agent

from .llm_setup import initialize_llm, TokenUsageCallbackHandler
from .db_utils import get_system_prompt
from .db_stats import log_api_call, log_agent_run
from .tools import collector_tools
from .agent_helpers import parse_json_output

logger = logging.getLogger(__name__)

MAX_SEARCH_ATTEMPTS = 2 # Reduced attempts as we now search per tag

# --- Main Collector Agent Function ---
def run_collector_agent(tags: list[str], search_depth: int = 10):
    """
    Runs the Collector agent to find new articles for the given tags.
    It performs a separate search for EACH tag.
    Args:
        search_depth: The number of search results to request from the google_search tool per tag.
        tags: A list of tags to search for.
    Returns:
        A list of dictionaries, where each dictionary represents a successfully scraped new article
        (e.g., {'url': '...', 'title': '...', 'content': '...'}).
        Returns an empty list if no new articles are found or an error occurs.
    """
    agent_name = "Collector"
    logger.info(f"{agent_name} Agent: Starting run for tags: {tags} (Search Depth: {search_depth})")
    if not tags:
        logger.warning(f"{agent_name} Agent: No tags provided. Skipping run.")
        return []

    start_time = time.time()
    run_success = False # Track success for logging
    attempt_count = 0
    all_collected_articles = [] # Accumulate results across all tags and attempts

    try:
        llm, token_callback = initialize_llm(provider='openai') # Get LLM and callback handler
        prompt_str = get_system_prompt(agent_name)
        if not prompt_str:
            logger.error(f"{agent_name} Agent: Failed to load system prompt.")
            return []

        # Define the prompt template - Updated to reflect per-tag search
        prompt = ChatPromptTemplate.from_messages([
            ("system", prompt_str),
            ("human", "Find relevant news articles for the specific tag: '{tag_query}'. Use the 'google_search' tool with num_results={search_depth}. Formulate your search query by prepending 'latest news' or 'recent news articles' to the tag. For example, if the tag is 'ai', search for 'latest news ai'. Use the available tools: 'google_search', 'find_rss_feed', 'parse_rss_feed', 'check_if_url_is_processed', 'scrape_web_content'. Follow the instructions precisely, especially regarding VALIDATION (skip homepages/category pages), checking the SYSTEM DOMAIN BLACKLIST, checking processed URLs BEFORE scraping, and returning the final result ONLY as a valid JSON list of objects, where each object has 'url', 'title', and 'content' keys."),
            # Langchain's agent executor needs this placeholder to insert intermediate steps
            ("placeholder", "{agent_scratchpad}"),
        ])
        # Create the agent
        collector_agent_runnable = create_openai_functions_agent(llm, collector_tools, prompt)

        # --- Retry Loop ---
        # Create the AgentExecutor once
        agent_executor = AgentExecutor(
            agent=collector_agent_runnable,
            tools=collector_tools,
            verbose=True, # Set to True for debugging agent steps
            handle_parsing_errors=True,
            max_iterations=15 # Max iterations per tag search
        )

        # --- Iterate through each tag ---
        for tag in tags:
            logger.info(f"{agent_name} Agent: Processing tag '{tag}'...")
            attempt_count = 0
            tag_collected_articles = [] # Articles found for this specific tag

            # --- Retry Loop per Tag ---
            while attempt_count < MAX_SEARCH_ATTEMPTS:
                attempt_count += 1
                logger.info(f"{agent_name} Agent: Search attempt {attempt_count}/{MAX_SEARCH_ATTEMPTS} for tag: '{tag}'")

                # Modify query slightly for retries to encourage different results
                if attempt_count == 1:
                    tag_query = tag # Agent will prepend 'latest news' based on prompt
                else:
                    # Explicitly add different phrasing for retry
                    tag_query = f"recent news articles about {tag}"
                    logger.info(f"{agent_name} Agent: Modified query for retry: '{tag_query}'")

                try:
                    # Reset token callback for each invocation if desired, or use cumulative
                    token_callback.reset() # Reset counts for this specific tag invocation

                    # Invoke the agent for the current tag
                    result = agent_executor.invoke({ # Input dictionary for the agent
                        "tag_query": tag_query, # Pass the single tag query
                        "search_depth": search_depth
                    }, config={"callbacks": [token_callback]}) # Pass callback handler to invoke

                    # Log the API call after invocation for this tag
                    log_api_call(
                        api_provider='OpenAI', # Assuming OpenAI for Collector
                        service_model=llm.model_name, # Get model name from LLM object
                        success=True, # Assuming success if invoke doesn't raise exception
                        caller_component=f"{agent_name} (Tag: {tag})",
                        input_tokens=token_callback.total_input_tokens, # Get tokens for this call
                        output_tokens=token_callback.total_output_tokens, # Get tokens for this call
                        metadata={'tag': tag, 'attempt': attempt_count}
                    )

                    # --- Output Parsing (Expecting JSON list) ---
                    raw_output = result.get('output', '')
                    parsed_articles = parse_json_output(
                        raw_output,
                        expected_keys=["url", "title", "content"],
                        agent_name=f"{agent_name} (Tag: {tag})"
                    )

                    # Accumulate valid articles found in this attempt for this tag
                    attempt_articles = parsed_articles if isinstance(parsed_articles, list) else []
                    tag_collected_articles.extend(attempt_articles)

                    if attempt_articles: # If this attempt found something, break retry loop for this tag
                        logger.info(f"{agent_name} Agent: Found {len(attempt_articles)} articles on attempt {attempt_count} for tag '{tag}'.")
                        break # Move to the next tag
                    elif attempt_count < MAX_SEARCH_ATTEMPTS:
                        logger.warning(f"{agent_name} Agent: Attempt {attempt_count} for tag '{tag}' resulted in zero valid articles. Retrying...")
                    else:
                        logger.warning(f"{agent_name} Agent: All {MAX_SEARCH_ATTEMPTS} attempts failed for tag '{tag}'.")

                except Exception as tag_error:
                     logger.error(f"{agent_name} Agent: Error during execution for tag '{tag}', attempt {attempt_count}: {tag_error}", exc_info=True)
                     # Log failed attempt
                     log_api_call(
                         api_provider='OpenAI',
                         service_model=getattr(llm, 'model_name', 'N/A'),
                         success=False,
                         caller_component=f"{agent_name} (Tag: {tag})",
                         input_tokens=token_callback.total_input_tokens,
                         output_tokens=token_callback.total_output_tokens,
                         metadata={'error': str(tag_error), 'tag': tag, 'attempt': attempt_count}
                     )
                     # Optionally break the retry loop for this tag on error, or continue retrying
                     # break # Let's break on error for now

            # Add articles found for this tag to the overall list
            all_collected_articles.extend(tag_collected_articles)
            logger.info(f"{agent_name} Agent: Finished processing tag '{tag}'. Found {len(tag_collected_articles)} articles for this tag.")
            # Add a small delay between processing tags to avoid rate limits if necessary
            # time.sleep(1)

        run_success = True # Mark as successful if loop completes without major error
        # The final return is handled by the finally block after deduplication

    except ValueError as ve:
        logger.error(f"{agent_name} Agent: Configuration error - {ve}")
        # Log failed attempt if LLM init failed
        log_api_call(api_provider='OpenAI', service_model='N/A', success=False, caller_component=agent_name, metadata={'error': str(ve), 'tags': tags})
        run_success = False
    except Exception as e: # Catch-all for unexpected errors during setup or looping
        logger.error(f"{agent_name} Agent: Critical error during execution for tags {tags}: {e}", exc_info=True)
        run_success = False
        # Logging for specific tag errors is handled inside the loop

    finally:
        end_time = time.time()
        # Deduplicate results based on URL before returning
        final_unique_articles = []
        seen_urls = set()
        for article in all_collected_articles:
            # Ensure article is a dict and has a URL before processing
            if isinstance(article, dict) and article.get('url'):
                if article['url'] not in seen_urls:
                    final_unique_articles.append(article)
                    seen_urls.add(article['url'])
            else:
                logger.warning(f"{agent_name}: Skipping invalid article data during final deduplication: {article}")

        log_agent_run(agent_name, start_time, end_time, run_success) # Log the overall agent run
        logger.info(f"{agent_name} Agent: Finished run for tags {tags}. Returning {len(final_unique_articles)} unique articles collected across all tags.")
        return final_unique_articles
